/*****************************************************************************/
/*  The leaf will be created in the X/Y plane bowed to Z. Base is in origin  */
/*  and it's tip points to +Y                                                */
/*  all parameters are in range 0..1                                         */
/*  max. size of the leaf may be defined as "lMaxLeafSize" and defaults to 5 */
/*  Ripple defines amplitude of the ripples on both borders                  */
/*  Weakness defines stiffness of the leaf. Weak leaves will be bowed to Z   */
/*  Age controlls texture only. Finish, pigment and normal will be influenced*/
/*  Tearfactor defines the probability of littel scratches no both borders.  */
/*  Scratches are created by omitting single triangles of the mesh           */
/*****************************************************************************/

// lLeafLOD influences the level of detail
// LOD settings
//      0 full detail
//      1 only single middle stem, reduced no. of triangles
//      2 additionally remove texture and again reduce no. of triangle
//      3 replace leaf by a single, flattened cone

#include "FernTools.inc"

// set defaults
#ifndef (lLeafSeed) #declare lLeafSeed = seed(123452); #end
#ifndef (lMaxLeafSize) #declare lMaxLeafSize = 5; #end
#ifndef (lLeafLOD) #declare lLeafLOD = 0; #end
#declare lLeafTextureCounter = -1;

// statistics
#ifndef (lTriangleCnt) #declare lTriangleCnt = 0; #end
#ifndef (lLeafCnt) #declare lLeafCnt = 0; #end

// smoothness of colour variances depending on age
#declare lLeafAgeSteps = 40;
#declare lLeafTextureLimit = 1000;

// all leaf textures are predefined and stored in an array. this is necessary for fast access and for the mesh triangles
#macro MakeLeafTextures()
        #if (lLeafLOD < 3)
                #debug "Building leaf textures ...\n"
                #declare lLeafBaseTextures = array[lLeafAgeSteps+1]

                #if (lLeafLOD = 0)        
                        #declare lLeafTextures = array[lLeafTextureLimit]
                        #declare lLeafTextureCounter = -1;
                #end
                
                #local lColours = array[lLeafAgeSteps+1]
        
                // normal definition
                #local lNormal = normal { bozo bump_size 0.03 scale 0.1 }
                #local lIllNormal = normal { bozo bump_size 0.2 scale 0.1 }
        
                // base textures
        
                #debug "   creating base textures...\n"         
                #local lAge = lLeafAgeSteps;
                #while (lAge >= 0 )
                        #local aAge = lAge/lLeafAgeSteps;
                
                        // compute colour according to age. defined in FernTools.inc
                        AgeToCol(aAge)

                        // pigment definition
                        // define base colour. slightly transparent and a small colour variance
                        #local lBaseCol  = color rgbf <lR,lG,lB,0.15>;
                        #local lLightCol = color rgbf <lR,lG+0.1,lB,0.15>;
                        #local lColours[lAge] = lBaseCol;
                
                        #local lPigment = pigment { bumps scale 0.0001+0.002*aAge
                                                    color_map { [0.0 lBaseCol]
                                                                [1.0 lLightCol] } }
                                                        
                        // finish definition. brilliance and highlights vary with age
                        #local lFinish = finish { diffuse 0.8 brilliance 0.4-0.4*pow(aAge,4) reflection 0  phong 0.3-0.3*aAge phong_size 2 }

                        #local lNormalHeight = 0.05*rand(lLeafSeed);
                        #local lNormal = normal { gradient y
                                        slope_map { [ 0.0 <  0, 1>]
                                                    [ 0.1 <lNormalHeight, 1>]
                                                    [ 0.1 <lNormalHeight,-1>]
                                                    [ 0.2 <  0,-1>]
                                                    [ 0.2 <  0, 0>]
                                                    [ 1.0 <  0, 0>] } scale 0.15}
                        #if (lLeafLOD = 0)
                                #declare lLeafBaseTextures[lAge] = texture { pigment { lPigment } finish { lFinish } normal { lNormal }}
                        #else
                                #declare lLeafBaseTextures[lAge] = texture { pigment { lPigment } finish { lFinish } }
                        #end
                
                        #local lAge = lAge -1;
                #end
        
                // stem pigment
                #local lStemPigment = pigment { rgb <0.62,0.33,0.16> }
                #local lStemFinish = finish { ambient 0 diffuse 0.7 reflection 0 phong 0.3 phong_size 10 }                 
                #declare lStemTexture = texture { pigment { lStemPigment } finish { lStemFinish } }
                #debug "Done\n"
        #end
#end

/*************************************************************************************/
/*                                                                                   */
/* Fern leaf                                                                         */
/*                                                                                   */
/*************************************************************************************/

// The leaf is oriented in the X/Y plane, tip to +Y, base in origin. weak leaves bow to +Z
// aSize                        0..1    [%] (of lMaxLeafSize) the leaf's size. NOTE: try to avoid 0, should be coordinated with aAge.
// aRippleFactor                0..1    [%] max. amplitude of the ripples. 0 gives a flat leaf, 1 a wrinkled one.
// aWeakNess                    0..1    [%] weakness. 0 is a stiff leaf (no Z extent), 1 a very weak one.
// aAge                         0..1    [%] the leaf's age. influences texture, "horizontal" wrinkles and proportions.
//                                              NOTE: will be split into discrete steps according to lLeafAgeSteps
// aTearFactor                  0..1    [%] probability of scratches and gaps in the leaf's surface
// aTranslate1                  <Vec>   translation vector to move the leaf's triangles before doing any other transformation
// aRotate1                     <Vec>   rotation vector 1. Leaf will be rotated in order Z->X->Y (!! Z first)
// aRotate2                     <Vec>   rotation vector 2. Additional rotation in standard order X->Y->Z
// aTranslate2                  <Vec>   translation vector to move the leaf's triangles to the final place (in relation to the wing)
#macro FernSubLeaf(aSize, aRippleFactor, aWeakness, aAge, aTearFactor,
                   aTranslate1, aRotate1, aRotate2, aTranslate2)

        // Finalisation macro to apply rotation and translations onto the triangles while computing them
        #macro Finalize(aVec)
                #declare lResult = vrotate(aVec+aTranslate1,<0,0,aRotate1.z>);
                #declare lResult = vrotate(lResult,<aRotate1.x,aRotate1.y,0>);
                #declare lResult = vrotate(lResult,aRotate2);
                lResult + aTranslate2
        #end
        
        #declare lLeafCnt = lLeafCnt + 1;
        
        // remap age to colour steps
        #local lAge = ceil(aAge*lLeafAgeSteps);
        
        // give a small random variation in the age
        #if ((lAge > 1) & (rand(lLeafSeed) > 0.5))
                #local lAge = lAge -1;
        #else 
                #if ((lAge < lLeafAgeSteps-1) & (rand(lLeafSeed) > 0.5)) #local lAge = lAge +1; #end
        #end
        
        // nr. of triangles depends on the leaf's size and the smoothness we need to show the small wrinkles
        #local lNumPoints = ceil(13+aRippleFactor*aSize*20);
        
        // if LOD > 0 reduce nr. of triangles
        #if ((lLeafLOD > 1) & (lNumPoints > 12)) #local lNumPoints = div(lNumPoints,2); #end

        // some arrays holding the data
        #ifdef (lCenter) #undef lCenter #end
        #ifdef (lBorderA) #undef lBorderA #end
        #ifdef (lBorderB) #undef lBorderB #end
        #ifdef (lNormalA) #undef lNormalA #end
        #ifdef (lNormalB) #undef lNormalB #end

        // due to a bug in POV we need a spare entry at the end, even if it is never used.
        #local lCenter = array[lNumPoints+1]
        #local lBorderA = array[lNumPoints+1]
        #local lBorderB = array[lNumPoints+1]
        #local lNormalA = array[lNumPoints+1]
        #local lNormalB = array[lNumPoints+1]
                                                           
        // compute the body of a single leaf. data is held in arrays
        // note: it computes both sides simultaneously
        #macro FernSubLeafBodyCalc(aSize, aRippleFactor, aWeakness)
        
                // local precalc
                #local lPointC = 0;        
                #local lRippleAngle1 = 3*rand(lLeafSeed);   // wrinkles right
                #local lRippleAngle2 = 3*rand(lLeafSeed);   // wrinkles left
                #local lRippleAngleC = 0;                   // stem shift (if any)
                #local lRippleAmp = 0;
                #local lRippleShift = aSize*0.4;
                #local lRippleStep = (60+20*rand(lLeafSeed))/lNumPoints; 
                #local lSizeY = lMaxLeafSize*aSize;
                #local lSizeX = lSizeY/(2+4.3*(1-exp(-aAge/0.2))); //lSizeY/(2+3*aAge)*(1-aAge*0.2);
                // compute triangles only for LOD < 3
                #while ((lPointC < lNumPoints) & (lLeafLOD < 3))
                        #local lAngle            = lPointC/lNumPoints*90;
                        #local lAngleRad         = radians(lAngle);
                        #local lAngleSin         = sin(lAngleRad);
                        #local lAngleCos         = cos(lAngleRad);
                
                        #local lAngle2           = lAngle*2;
                        #local lAngle2Rad        = radians(lAngle2);
                
                        #local lVariance         = sin(lAngle2Rad);
                        #local lY                = lSizeY*lAngleSin;
                        #local lX                = lSizeX*0.5*aAge*aAge*sin(lRippleAngleC);
                        #local lXBase            = lSizeX*lAngleCos;
                        #local lX1               = lX+  lXBase+lSizeX*0.05*lVariance*sin(lRippleAngle1*2);
                        #local lX2               = lX- (lXBase+lSizeX*0.05*lVariance*sin(lRippleAngle2*2));
                        #local lRippleStep1      = lRippleStep*(0.95+0.1*rand(lLeafSeed));
                        #local lRippleStep2      = lRippleStep*(0.95+0.1*rand(lLeafSeed));
                        #local lRippleStepC      = lRippleStep*aAge*0.2*(0.6+0.4*rand(lLeafSeed));
                        
                        // stem
                        #declare lCenter[lPointC] = <lX,lY,lSizeY*aWeakness*(1-lAngleCos)>;
                        #local lZBase = lCenter[lPointC].z -(0.07 +aSize*0.25)*lAngleCos;
                        #local lZMult = lRippleAmp*lAngleCos;
                                
                        #local lZ1 = lZBase + lZMult*sin(lRippleAngle1) - lRippleShift*lVariance;
                        #local lZ2 = lZBase + lZMult*sin(lRippleAngle2) - lRippleShift*lVariance;
                        
                        // borders (take care about gaps)
                        #if (rand(lLeafSeed) < aTearFactor)
                                #declare lBorderA[lPointC] = <lCenter[lPointC].x+0.01,lY,lCenter[lPointC].z+0.01>;
                        #else
                                #declare lBorderA[lPointC] = <lX1,lY,lZ1>;
                        #end
                
                        #if (rand(lLeafSeed) < aTearFactor)
                                #declare lBorderB[lPointC] = <lCenter[lPointC].x-0.01,lY,lCenter[lPointC].z+0.01>;
                        #else
                                #declare lBorderB[lPointC] = <lX2,lY,lZ2>;
                        #end
                        
                        // defaults. will be overwritten
                        #declare lNormalA[lPointC] = <0,0,-1>;
                        #declare lNormalB[lPointC] = <0,0,-1>;

                        // some hand dressed softening of the left, lower edge
                        #if (lPointC = 0) #declare lBorderB[lPointC] = <lX2*0.3,0,lZ2*0.3>; #end
                        #if (lPointC = 1) #declare lBorderB[lPointC] = <lX2*0.8,0.07,lZ2*0.8>; #end
                        #if (lPointC = 2) #declare lBorderB[lPointC] = <lX2*0.9,0.2,lZ2*0.9>; #end
                        
                        // compute ripples for next step                                
                        #local lRippleAngle1 = lRippleAngle1 + lRippleStep1*rand(lLeafSeed);
                        #local lRippleAngle2 = lRippleAngle2 + lRippleStep2*rand(lLeafSeed);
                        #local lRippleAngleC = lRippleAngleC + lRippleStepC*rand(lLeafSeed);
                        #local lRippleAmp = aSize*0.3*aRippleFactor*lVariance;
                        
                        #local lPointC = lPointC +1;
                #end
                
                // smoothen the surface since we use as less triangles as possible
                CalcNormals(lBorderA, lNormalA, 2, lNumPoints-1)
                CalcNormals(lBorderB, lNormalB, 2, lNumPoints-1)
                
                // finalize the points and their normals
                #local lPointC = 0;
                #while (lPointC < lNumPoints)
                        #declare lBorderA[lPointC] = Finalize(lBorderA[lPointC]);
                        #declare lBorderB[lPointC] = Finalize(lBorderB[lPointC]);
                        #declare lCenter[lPointC] = Finalize(lCenter[lPointC]);
                        #declare lNormalA[lPointC] = Finalize(lNormalA[lPointC])-aTranslate2;
                        #declare lNormalB[lPointC] = Finalize(lNormalB[lPointC])-aTranslate2;
                        #local lPointC = lPointC + 1;
                #end
        #end
        
        // this builds the mesh according to the computed data
        #macro FernSubLeafBody(aBorder, aNormals, aSide)
                #local lSide = 1-2*mod(aSide,2);
                #local lMiddleNormal = Finalize(< -0.5*lSide,0,-1>)-aTranslate2;         // normal of the edges at the stem
//                #local lStemOffs     = Finalize(<0.02*lSide*aAge,0,0>)-aTranslate2;           // offset of the leaf from the stem
                
                #if (lLeafLOD > 0)
                        // LOD 1 and 2: no normals, just plain texture
                        #local lTextIdx = aSide;
                        #local lTextStr = concat("lLeafBaseTextures[",str(lAge,2,0),"]")         // get texture for the leaf
                        #local lText = lLeafBaseTextures[lAge]
                #else
                        #declare lLeafTextureCounter = lLeafTextureCounter+1;
                        #local lTextIdx = lLeafTextureCounter;
                        #declare lLeafTextures[lLeafTextureCounter] = texture { lLeafBaseTextures[lAge]
                                                                                 rotate z*45*lSide
                                                                                 rotate y*30*lSide
                                                                                 rotate z*aRotate1.z
                                                                                 rotate <aRotate1.x,aRotate1.y,0>
                                                                                 rotate aRotate2  }
                        #local lTextStr = concat("lLeafTextures[",str(lTextIdx,1,0),"]")
                        #local lText = lLeafTextures[lTextIdx]
                        #ifdef (lTextureFile)
                                #write (lTextureFile, "#declare lLeafTextures[",str(lTextIdx,1,0),"]=texture{lLeafBaseTextures[",str(lAge,1,0),
                                                      "] rotate z*",str(45*lSide,1,0),
                                                       " rotate y*",str(30*lSide,1,0),
                                                       " rotate z*",str(aRotate1.z,7,4),
                                                       " rotate <",str(aRotate1.x,7,4),",",str(aRotate1.y,7,4),",0> ",
                                                       " rotate <",str(aRotate2.x,7,4),",",str(aRotate2.y,7,4),",",str(aRotate2.z,7,4),">}\n")
                        #end
                #end
                
                #local lPoint = 0;
                #while (lPoint < lNumPoints-1)
                        #local lNextPoint = lPoint+1;
                        #local lC1 = lCenter[lPoint]; //+ lStemOffs;
                        #local lC2 = lCenter[lNextPoint]; //+ lStemOffs;

                        PutTriangle ( aBorder[lPoint]     ,aNormals[lPoint],
                                      aBorder[lNextPoint] ,aNormals[lNextPoint],
                                      lC1                 ,lMiddleNormal,
                                      lText,lTextStr )

                        PutTriangle( lC1                  ,lMiddleNormal,
                                     lC2                  ,lMiddleNormal, 
                                     aBorder[lNextPoint]  ,aNormals[lNextPoint],
                                     lText,lTextStr )

                        #local lPoint = lPoint +1;
                #end
        #end

        // build the stem
        #macro FernSubLeafBodyStem()
                #local lStemPoints = 6-lLeafLOD;
        
                #local lStem = array[lStemPoints+1]
                #local lLastStem = array[lStemPoints+1]
                #local lStemNormal = array[lStemPoints+1]
                #local lLastStemNormal = array[lStemPoints+1]
        
                #local lEndK = 2*pi;
        
                #local lTextStr = "lStemTexture"
                
                #local lPoint = 0;
                #local lLastLeafStemPoint = -1;
        
                #if (lLeafLOD = 0)
                        #local lPointStep = 1; // max(1,ceil(0.1*lNumPoints/aSize));
                #else
                        #local lPointStep = max(1,ceil(0.18*lNumPoints/aSize)/(pow((aAge+aWeakness)/2,0.33333)+0.01));
                #end
        
                #while (lPoint < lNumPoints)
                        #local lAmp = 0.01+(0.08*(1-lPoint/lNumPoints));
                        #if (lLastLeafStemPoint >= 0)
                                #local lDiff = vnormalize(lCenter[lPoint]-lCenter[lLastLeafStemPoint]);
                                #if (vlength(lDiff) = 0)
                                        #local lDiff = <0,1,0>;
                                #end
                        #else
                                #local lDiff = <0,1,0>;
                        #end
                
                        #local lk = 0;
                        #while (lk < lStemPoints)
                                #local lFactor = lk/lStemPoints;
                                #if (lLeafLOD = 0)
                                        #local lX = lAmp+lAmp*1.2*sin(lFactor*lEndK);
                                #else 
                                        #local lX = lAmp*1.5*sin(lFactor*lEndK);
                                #end
        
                                #local lZ = lAmp*0.8*cos(lFactor*lEndK);
                        
                                #local lStem[lk] = Finalize(<lX,0,lZ>)-aTranslate2;
                                
                                #local lk = lk+1;
                        #end
                        
                        // calc normals
                        #local lk = 0;
                        #while (lk < lStemPoints)
                                #local lNextK = mod(lk+1,lStemPoints);
                                #local lNextNextK = mod(lk+2,lStemPoints);
                                
                                #local lV = vnormalize(lStem[lNextNextK]-lStem[lk]); 
                                #local lStemNormal[lNextK] = Finalize(vcross(lDiff,lV))-aTranslate2;
                                #local lk = lk+1;
                        #end
                        
                        #if (lLastLeafStemPoint >= 0)
                                #local lHighEnd = lCenter[lPoint];
                                #local lLowEnd = lCenter[lLastLeafStemPoint];
                        #end
                
                        #local lk = 0;
                        #while (lk < lStemPoints)   
                                #local lNextK = mod(lk+1,lStemPoints);
                                #if (lLastLeafStemPoint >= 0)
                                        PutTriangle ( lHighEnd+lStem[lk],               lStemNormal[lk],
                                                      lHighEnd+lStem[lNextK],           lStemNormal[lNextK],
                                                      lLowEnd+lLastStem[lk],            lLastStemNormal[lk],
                                                      lStemTexture,lTextStr )
                                        PutTriangle ( lLowEnd+lLastStem[lk],            lLastStemNormal[lk],
                                                      lLowEnd+lLastStem[lNextK],        lLastStemNormal[lNextK],
                                                      lHighEnd+lStem[lNextK],           lStemNormal[lNextK],
                                                      lStemTexture,lTextStr )

                                        #if (lLeafLOD = 0)
                                                PutTriangle ( lHighEnd-lStem[lk],          lStemNormal[lk],
                                                              lHighEnd-lStem[lNextK],      lStemNormal[lNextK],
                                                              lLowEnd-lLastStem[lk],       lLastStemNormal[lk],
                                                              lStemTexture,lTextStr )
                                                PutTriangle ( lLowEnd-lLastStem[lk],       lLastStemNormal[lk],
                                                              lLowEnd-lLastStem[lNextK],   lLastStemNormal[lNextK],
                                                              lHighEnd-lStem[lNextK],      lStemNormal[lNextK],
                                                              lStemTexture,lTextStr )
                                        #end
                                #end
                                #local lk = lk+1;
                        #end
                
                        #local lk = 0;
                        #while (lk < lStemPoints)
                                #local lLastStem[lk] = lStem[lk];
                                #local lLastStemNormal[lk] = lStemNormal[lk];
                                #local lk = lk+1;
                        #end
                
                        #local lLastLeafStemPoint = lPoint;
                
                        // ensure that last point is drawn regardless of the step size
                        #if ((lPoint + lPointStep) >= lNumPoints)
                                #if (lPoint < lNumPoints-1)
                                        #local lPoint = lNumPoints-1;
                                #else
                                        #local lPoint = lNumPoints;
                                #end
                        #else
                                #local lPoint = lPoint + lPointStep;
                        #end
                #end
        #end

        // build the leaf
        #if (lLeafLOD < 3)
                // compute the mesh                
                FernSubLeafBodyCalc(aSize,aRippleFactor,aWeakness)
                
                // right side mesh (A)        
                FernSubLeafBody(lBorderA, lNormalA, 0)

                // left side mesh (B)        
                FernSubLeafBody(lBorderB, lNormalB, 1)
                        
                // stem
                FernSubLeafBodyStem()
        #else
                cone { <0,0,0>,lMaxLeafSize*aSize/(2+4.3*(1-exp(-aAge/0.2)))
                //aSize/(2+3*aAge)*(1-aAge*0.2),
                <0,lMaxLeafSize*aSize,0>,0
                       scale <1,1,0.01>
                       
                #if (lLeafLOD > 2)
                        AgeToCol(0)
                        pigment { rgb <lR,lG,lB> }
                        finish { ambient 0 diffuse 1 }
                #else
                       texture { lLeafTextures[ceil(aAge*lLeafAgeSteps)][0] } 
                #end
                       rotate x*45/2*aWeakness
                       
                       translate aTranslate1
                       rotate <0,0,aRotate1.z>
                       rotate <aRotate1.x,aRotate1.y,0>
                       rotate aRotate2
                       translate aTranslate2
                }
        #end
#end
