/******************************************************************************/
/** orientation of the wing depends on the mode. base is always in origin    **/
/** and the wing "grows" upwards (to +Y). as the wing gets larger and        **/
/** therefore "heavier" it bows more and more to the +Z plane                **/
/** mode 0 for hanging ferns: depending on age it will hang downwards to -Y  **/
/** mode 1 for standing ferns: depending on age it will extend more and more **/
/**     to +Z, will never go beyond Y=0. very old wings will end up at Y=0.  **/
/******************************************************************************/

// get the leaf macros
#include "FernSubLeaf.inc"

// define defaults
#ifndef (lWingSeed) #declare lWingSeed = seed(52357); #end
#ifndef (lDiedSeed) #declare lDiedSeed = seed(23423); #end
#ifndef (lWingLOD) #declare lWingLOD = 0; #end
#ifndef (lLeafLOD) #declare lLeafLOD = 0; #end
                 
// statistics
#ifndef (lWingCnt) #declare lWingCnt = 0; #end

// nr. of step used for the stem of the wing. increas this for smoother bows in the wing.
#declare lStemColSteps = 20;

// auto LOD support. changes leaf LOD depending on the distance to a specific point
#macro AutoLOD(aVec, aSize)
        #ifndef (lFernLODScale) #declare lFernLODScale = 1; #end
        // compute LeafLOD automatically according to the position of the leaf
        
        #local lDist = vlength(lFernLODCenter-aVec)/lFernLODScale*(1+aSize);
        
        #if (lDist < 45) 
                #declare lBaseLeafLOD = 0;
        #else 
                #if (lDist < 90) 
                        #declare lBaseLeafLOD = 1;
                #else 
                        #declare lBaseLeafLOD = 2;
                #end 
        #end
        #if (lWingLOD > 0)
                #declare lBaseLeafLOD = lBaseLeafLOD + 4;
        #end
#end

#macro MakeWingTextures()
        #debug "Building stem textures ...\n"
        #declare lWingTextures = array[lStemColSteps+1]
        #local lStep = 0;
        #while (lStep <= lStemColSteps)
                #local aFactor = lStep/lStemColSteps;
                #if (lWingLOD = 0)                        
                        AgeToCol(aFactor)
                        #declare lWingTextures[lStep] = texture { pigment { rgb <lR,lG,lB> }
                                              finish { ambient 0.1 diffuse 0.8 reflection 0 }
                                              normal { bumps bump_size 0.5 scale 0.005 } }
                #else
                        AgeToCol(aFactor)
                        #declare lWingTextures[lStep] = texture { pigment { rgb <lR,lG,lB> }
                                                                  finish { ambient 0.4 diffuse 1 } }
                #end
                #local lStep = lStep +1;
        #end
        #debug "Done\n"
#end

#macro FernWing(aStyle,         // 0=hanging, 1=standing
                aAge,           // 0..1 the wings's age
                aWideNess,      // 1.. any number horizontal stretching                
                aDiedFactor,    // 0..1 % of dead leaves on this wing 
                aLostFactor,    // 0..1 % of lost (missing) leaves
                aWindingFactor, // 0..1 % of curveness
                aShapeA, aShapeB)   // shape control parameters (see fernshape.xls)
                         
        // check textures
        #ifndef (lWingTextures)
                MakeWingTextures()
        #end
        #ifndef (lLeafTextures)
                MakeLeafTextures()
        #end
        
        // prepare output to file (always level 0!)
        #ifdef (lOutFile)
                #declare lWingLOD = 0;
                #declare lLeafLOD = min(2,lLeafLOD);  // default may be overwritten by LODCenter
                #write (lOutFile,"mesh{\n")
        #end
        
        // standing fern has minimum wing age of 0.2
        #if (aStyle = 1)
                #local lAge = 0.2+0.8*aAge;
        #else
                #local lAge = aAge;
        #end
        
        // length of stem without leaves is always in size og 3 leaves
        #local lSkippedLeaves = 3;
        
        // count wing for debug output
        #declare lWingCnt = lWingCnt +1; 
        
        // no died leaves in preview wings! (would disturb in auto LOD)
        #if (lWingLOD > 0 ) #local aDiedFactor = 0; #end
        
        // compute stem
        #local lLeafCenter = array[200]
        #local lWingLastPoint = <0,0,0>;
        #local lDelta = 0;
        #local lDist = 0;
        #local lNumLeaves = -1;
        #local lWind = (-10+5*rand(lWingSeed))*aWindingFactor;
        #local lFactor = 0;
        #if (lWingLOD < 2)
                #local lStep = 1/200;
        #else
                #local lStep = 1/(30+(30*(1-lAge)));
        #end
        
        #debug concat("FernWing (",str(lWingCnt,3,0),") computing stem...\n")
        #local lAge2 = lAge*lAge;        
        #local lIdx = 0;
        #while (lFactor <= 1)
                #if (aStyle = 0)
                        // hanging style
                        #local lMaxAngle = pi*(lAge2/2.5+lAge*0.4+0.2);
                        #local lAmp1 = (50-30*lAge2)*aWideNess;
                        #local lAmp2 = 25*cos(lAge*pi/2)+21;
                        #local lAmp3 = 35/(1+exp(-2*(lAge-0.5)/0.3))+15;
                
                        #local lZ = lAmp1*(1-(0.5*cos(lMaxAngle*lFactor)+0.5));
                        #local lY = lAmp2*sin(lMaxAngle*lFactor)-lAmp3*lFactor;
                #else
                        // standing style
                        #local lZMax = 50*lAge2*lAge;
                        #local lYMax = 5+15*pow(sin(lAge*pi),2);
                        #local lExpDiv = 1-0.75*lAge2;
                        #local lExp = (1-exp(-lFactor/lExpDiv))*pi*lAge;
                        
                        #local lZ = lZMax*lFactor;
                        #local lY = lYMax*sin(lExp);
                #end
                #local lX = lWind*(1-lFactor)*sin(lFactor*2*pi);
                
                #if (lIdx > 0)
                        #local lDist = vlength(<lX,lY,lZ> - lWingLastPoint);
                        #local lDelta = lDelta + lDist;
                        #if (lDelta >= 2-(1.5*pow(lFactor,4)))
                                #local lDelta = 0;
                                #local lNumLeaves = lNumLeaves + 1;
                                #if (lNumLeaves >=0)
                                        #local lLeafCenter[lNumLeaves] = <lX,lY,lZ>;
                                #end
                        #end
                #end
                #local lWingLastPoint = <lX,lY,lZ>;                
                #local lIdx = lIdx +1;                
                #local lFactor = lFactor + lStep;
        #end
        
        // since length is known now, we know how many leaves will fit onto the stem
        #local aNumLeaves = lNumLeaves - lSkippedLeaves;

        // in case of mixed mesh/union due to leaf-LOD = 3 we have to combine this by using a union
        #if ((lWingLOD < 2) & (lLeafLOD >2))
                union {
        #end
        
        // main stem itself
        #debug concat("FernWing (",str(lWingCnt,3,0),") building stem...\n")
        mesh {
                #local lStemSteps = 6;                
                #local lLastStep = 0;
                #local lStep = 1;                
                
                #while (lStep < lNumLeaves)
                        #local lFactor = (1-lLastStep/lNumLeaves);
                        #local lIntFactor = ceil(lFactor*lStemColSteps);
                        #local lTip = 0.02+0.2*lFactor;
                        #local lBase = 0.02+0.2*(1-lStep/lNumLeaves);
                        #if (lWingLOD > 1)
                                #local lBase = lBase *3;
                                #local lTip = lTip *3;
                        #end
                        
                        #local lV = lLeafCenter[lLastStep]-lLeafCenter[lStep];
                        #local lXRot = degrees(atan2(lV.z,lV.y));
                        #local lYRot = degrees(atan2(lV.x,lV.z));
                        
                        #local lSinArray = array[lStemSteps]                        
                        #local lCosArray = array[lStemSteps]                        
                        #local lSubStep = 0;
                        #while (lSubStep < lStemSteps)
                                #local lAngle = lSubStep/lStemSteps*2*pi;
                                #local lSinArray[lSubStep] = sin(lAngle);
                                #local lCosArray[lSubStep] = cos(lAngle);
                                #local lSubStep = lSubStep + 1;
                        #end
                        
                        #local lWingTexture = lWingTextures[lIntFactor]
                        #local lTextStr = concat("lWingTextures[",str(lIntFactor,2,0),"]")
                        
                        #local lSubStep = 0;                        
                        #while (lSubStep <= lStemSteps)
                                #local lMStep = mod(lSubStep,lStemSteps);
                                #local lSin = lSinArray[lMStep];
                                #local lCos = lCosArray[lMStep];
                                #local lSinVec = <lBase*lSin,0,lBase*lCos>;
                                #local lCosVec = <lTip*lSin,0,lTip*lCos>;
                                
                                #local lVecBase = vrotate(lSinVec, <lXRot,lYRot,0>)+lLeafCenter[lStep];
                                #local lVecTip = vrotate(lCosVec, <lXRot, lYRot,0>)+lLeafCenter[lLastStep];
                                
                                #if (lSubStep > 0)
                                        PutTriangle( lLastBase, lLastSinVec,
                                                     lVecBase, lSinVec,
                                                     lVecTip, lCosVec, lWingTexture,lTextStr)
                                        PutTriangle( lLastTip, lLastCosVec,
                                                     lVecTip, lCosVec,
                                                     lLastBase, lLastSinVec, lWingTexture,lTextStr)
                                #end
                                
                                #local lLastBase = lVecBase;                                
                                #local lLastTip = lVecTip;
                                #local lLastSinVec = lSinVec;
                                #local lLastCosVec = lCosVec;
                                #local lSubStep = lSubStep + 1;
                        #end
                        
                        #local lLastStep = lStep;
                        #local lStep = lStep + 1;
                #end
        #debug concat("FernWing (",str(lWingCnt,3,0),") building leaves...\n")
        
        #if (lWingLOD < 2 )
                // now add the leaves (LOD 0 and 1 only)
                
                #if (lLeafLOD > 2) // if LOD of leaves is 0,1, or 2 they will be a mesh. otherwise a list of cones!
                        }
                        union {
                #end
                
                #local lLeaf = lSkippedLeaves;
                #while (lLeaf < lNumLeaves-1)
        
                        #local lStemOffset = (0.02+0.2*(1-lLeaf/lNumLeaves));
                        
                        // compute rotation angles according to the main stem
                        #local lV = lLeafCenter[lLeaf+1]-lLeafCenter[lLeaf];
                        #local lXRot = degrees(atan2(lV.z,lV.y));
                        #local lYRot = degrees(atan2(lV.x,lV.z));
                        
                        #local lFactor          = (lLeaf-lSkippedLeaves)/aNumLeaves;
                        #local lAngle           = lFactor*270;
                        #local lAngleRad        = radians(lAngle);
                        #local lExpMul          = aShapeA;
                        #local lExpDiv          = aShapeB;
                        #local lExp             = lExpMul*exp(-lFactor/lExpDiv);
                        #local lSize            = (sin(lAngleRad*lExp)+0.05)*(0.95+0.1*rand(lWingSeed));
                        
                        // compute leaf shapes and ages
                        #local lAge             = 0.76*exp(-lFactor/0.3);//exp(-lFactor/0.7);
                        #local lRipple          = (cos(radians(lFactor*180))*0.5+0.5)*(0.9+0.2*rand(lWingSeed));
                        #local lWeak            = exp(-lFactor/0.4)*(0.9+0.2*rand(lWingSeed));
                        
                        #ifdef (lFernLODCenter)
                                AutoLOD(lLeafCenter[lLeaf], lSize)
                        #else
                                #declare lBaseLeafLOD = lLeafLOD;
                        #end
                        
                        #if (rand(lDiedSeed) >= aLostFactor)
                                #if (rand(lDiedSeed) < aDiedFactor)
                                        #local lAge    = 0.85+0.15*rand(lLeafSeed);
                                        #local lRipple = 1;
                                        #local lWeak   = 1;
                                        #local lSize = lSize / 2;
                                        #declare lLeafLOD = 0;
                                #else
                                        #declare lLeafLOD = lBaseLeafLOD;
                                #end
                        
                                FernSubLeaf(lSize,lRipple,lWeak,lAge,0.01,  
                                                <0,lStemOffset,0>,
                                                <20,30+(20*rand(lWingSeed)-10),-90>,
                                                <lXRot, lYRot, 0>,
                                                lLeafCenter[lLeaf]) //+<0.2,0,0>)
                        #end
                        
                        #if (rand(lDiedSeed) >= aLostFactor)
                                #if (rand(lDiedSeed) < aDiedFactor)
                                        #local lAge    = 0.85+0.15*rand(lLeafSeed);
                                        #local lRipple = 1;
                                        #local lWeak   = 1;
                                        #local lSize = lSize / 2;
                                        #declare lLeafLOD = 0;
                                #else
                                        #declare lLeafLOD = lBaseLeafLOD;
                                #end

                                FernSubLeaf(lSize,lRipple,lWeak,lAge,0.01,
                                                <0,lStemOffset,0>,
                                                <20,-(30+(20*rand(lWingSeed)-10)),90>,
                                                <lXRot, lYRot, 0>,
                                                (<(lLeafCenter[lLeaf].x+lLeafCenter[lLeaf+1].x)/2,
                                                  (lLeafCenter[lLeaf].y+lLeafCenter[lLeaf+1].y)/2,
                                                  (lLeafCenter[lLeaf].z+lLeafCenter[lLeaf+1].z)/2>))//-<0.2,0,0>))
                        #end
                        
                        #debug concat("FernWing (",str(lWingCnt,3,0),")",str(lLeaf/lNumLeaves*100,3,0),"% done...\n")
                        #declare lLeaf = lLeaf +1;
                #end
        #end
        } // close mesh (or union in case of leaf-LOD = 3)
        
        // close union if leaves on wing and leaf-LOD = 3
        #if ((lWingLOD < 2) & (lLeafLOD >2))
                }  // of outermost union
        #end

        #debug concat("FernWing (",str(lWingCnt,3,0),") completed.","\n\n")
        #ifdef (lOutFile)
                #write (lOutFile,"}\n")
        #end

#end
